/***************************************************************************
 $RCSfile: account.cpp,v $
                             -------------------
    cvs         : $Id: account.cpp,v 1.18 2005/03/29 02:02:02 aquamaniac Exp $
    begin       : Mon Mar 01 2004
    copyright   : (C) 2004 by Martin Preuss
    email       : martin@libchipcard.de

 ***************************************************************************
 *          Please see toplevel file COPYING for license details           *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "account.h"
#include "app.h"

#include <gwenhywfar/debug.h>
#include <stdlib.h>
#include <assert.h>



Account::Account(App *app)
:_accountStatus(0)
,_bankingId(0)
,_banking(app)
,_storage(0)
,_lastTransactionDate(0) {
  assert(app);

}



Account::Account(App *app, const AB_ACCOUNT *a)
:_accountStatus(0)
,_bankingId(0)
,_banking(app)
,_storage(0)
,_lastTransactionDate(0){
  const char *p;

  assert(a);
  p=AB_Account_GetCountry(a);
  if (p)
    _country=p;
  p=AB_Account_GetAccountNumber(a);
  if (p)
    _accountNumber=p;
  p=AB_Account_GetAccountName(a);
  if (p)
    _accountName=p;
  p=AB_Account_GetBankCode(a);
  if (p)
    _bankCode=p;
  p=AB_Account_GetBankName(a);
  if (p)
    _bankName=p;
  p=AB_Account_GetOwnerName(a);
  if (p)
    _ownerName=p;
  _bankingId=AB_Account_GetUniqueId(a);
}


Account::Account(const Account &a)
:_accountStatus(0)
,_bankingId(a._bankingId)
,_banking(0)
,_storage(0){
  _bankCode=a._bankCode;
  _accountNumber=a._accountNumber;
  _bankName=a._bankName;
  _accountName=a._accountName;
  _ownerName=a._ownerName;
  _transactions=a._transactions;
  _bankingId=a._bankingId;
  _banking=a._banking;
  _lastTransactionDate=a._lastTransactionDate;
  if (a._accountStatus)
    _accountStatus=AB_AccountStatus_dup(a._accountStatus);

}


Account::~Account(){
  AB_AccountStatus_free(_accountStatus);
}



bool Account::fromDb(GWEN_DB_NODE *db) {
  GWEN_DB_NODE *dbT;

  _country=GWEN_DB_GetCharValue(db, "country", 0, "de");
  _bankCode=GWEN_DB_GetCharValue(db, "bankCode", 0, "");
  _accountNumber=GWEN_DB_GetCharValue(db, "accountNumber", 0, "");
  _bankName=GWEN_DB_GetCharValue(db, "bankName", 0, "");
  _accountName=GWEN_DB_GetCharValue(db, "accountName", 0, "");
  _ownerName=GWEN_DB_GetCharValue(db, "ownerName", 0, "");
  dbT=GWEN_DB_GetGroup(db, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "accountStatus");
  if (dbT)
    _accountStatus=AB_AccountStatus_fromDb(dbT);
  _bankingId=GWEN_DB_GetIntValue(db, "bankingId", 0, 0);
  dbT=GWEN_DB_GetGroup(db, GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                       "lastTransactionDate");
  if (dbT)
    _lastTransactionDate=GWEN_Time_fromDb(dbT);
  DBG_INFO(0, "Loaded account from DB (%s/%s)",
           _bankCode.c_str(),
           _accountNumber.c_str());
  return true;
}



bool Account::toDb(GWEN_DB_NODE *db) {
  if (!_country.empty())
    GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                         "country", _country.c_str());
  if (!_bankCode.empty())
    GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                         "bankCode", _bankCode.c_str());
  if (!_accountNumber.empty())
    GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                         "accountNumber", _accountNumber.c_str());
  if (!_bankName.empty())
    GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                         "bankName", _bankName.c_str());
  if (!_accountName.empty())
    GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                         "accountName", _accountName.c_str());
  if (!_ownerName.empty())
    GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                         "ownerName", _ownerName.c_str());
  GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                      "bankingId", _bankingId);

  if (_accountStatus) {
    GWEN_DB_NODE *dbT;

    dbT=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_DEFAULT, "accountStatus");
    assert(dbT);
    if (AB_AccountStatus_toDb(_accountStatus, dbT)) {
      DBG_ERROR(0, "Could not store account status");
      return false;
    }
  }

  if (_lastTransactionDate) {
    GWEN_DB_NODE *dbT;

    dbT=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_DEFAULT, "lastTransactionDate");
    assert(dbT);
    if (GWEN_Time_toDb(_lastTransactionDate, dbT)) {
      DBG_ERROR(0, "Could not store date of last transaction");
      return false;
    }
  }

  return true;
}



const std::string &Account::getCountry() const{
  return _country;
}



void Account::setCountry(const std::string &s){
  _country=s;
}



const std::string &Account::getBankCode() const{
  return _bankCode;
}



void Account::setBankCode(const std::string &s){
  _bankCode=s;
}



const std::string &Account::getAccountNumber() const{
  return _accountNumber;
}



void Account::setAccountNumber(const std::string &s){
  _accountNumber=s;
}



const std::string &Account::getBankName() const{
  return _bankName;
}



void Account::setBankName(const std::string &s){
  _bankName=s;
}



const std::string &Account::getAccountName() const{
  return _accountName;
}



void Account::setAccountName(const std::string &s){
  _accountName=s;
}



const std::string &Account::getOwnerName() const{
  return _ownerName;
}



void Account::setOwnerName(const std::string &s){
  _ownerName=s;
}



GWEN_TYPE_UINT32 Account::getBankingId() const{
  return _bankingId;
}



void Account::setBankingId(GWEN_TYPE_UINT32 i){
  _bankingId=i;
}



const AB_ACCOUNT_STATUS *Account::getAccountStatus() const{
  return _accountStatus;
}



void Account::setAccountStatus(const AB_ACCOUNT_STATUS *s){
  assert(s);
  _accountStatus=AB_AccountStatus_dup(s);;
}



const std::list<RefPointer<Transaction> >  &Account::getTransactions() {
  if (_transactions.size()==0) {
    if (!_loadTransactions()) {
      DBG_ERROR(0, "Could not load transactions");
    }
  }
  return _transactions;
}



bool Account::_loadTransactions(unsigned int fromYear,
                                unsigned int fromMonth,
                                unsigned int fromDay) {
  int year, month, day;
  int rv;
  GWEN_TYPE_UINT32 ymd;
  GWEN_IDLIST *idl;

  DBG_NOTICE(0, "Loading transactions");
  _transactions.clear();
  ymd=(fromYear<<16)+(fromMonth<<8)+fromDay;
  if (!getStorage())
    return false;

  idl=AH_Storage_GetAvailableDays(_storage);
  if (!idl) {
    DBG_ERROR(0, "Error building storage index");
    return false;
  }
  rv=AH_Storage_GetFirstDay(_storage, idl, &year, &month, &day);
  while(!rv) {
    GWEN_TYPE_UINT32 id;
    GWEN_DB_NODE *db;
    GWEN_TYPE_UINT32 cymd;

    DBG_DEBUG(0, "Loading day %04d/%02d/%02d",
              year, month, day);
    cymd=(year<<16)+(month<<8)+day;
    if (cymd>=ymd) {
      id=AH_Storage_OpenDay(_storage, year, month, day, 1);
      if (!id) {
        DBG_ERROR(0, "Error loading day %04d/%02d/%02d",
                  year, month, day);
        GWEN_IdList_free(idl);
        return false;
      }

      db=AH_Storage_GetFirstTransaction(_storage, id);
      if (!db) {
        DBG_WARN(0, "No transactions in day %04d/%02d/%02d",
                 year, month, day);
      }
      while(db) {
        RefPointer<Transaction> t;

        if (!t.ref().fromDb(db)) {
          DBG_ERROR(0, "Bad format of transaction in day %04d/%02d/%02d",
                    year, month, day);
          AH_Storage_AbandonDay(_storage, id);
          GWEN_IdList_free(idl);
          return false;
        }
        DBG_ERROR(0, "Got a transaction");
        _transactions.push_back(t);
        db=AH_Storage_GetNextTransaction(_storage, id);
      }
      if (AH_Storage_CloseDay(_storage, id)) {
        DBG_ERROR(0, "Error closing day %04d/%02d/%02d",
                  year, month, day);
        GWEN_IdList_free(idl);
        return false;
      }
    }
    rv=AH_Storage_GetNextDay(_storage, idl, &year, &month, &day);
    if (rv) {
      DBG_NOTICE(0, "End of day list.");
    }
  } /* if first day */
  GWEN_IdList_free(idl);

  return true;
}



bool Account::loadTransactions(GWEN_TIME *fromDate,
                               GWEN_TIME *toDate,
                               std::list<Transaction> &tl) {
  int year, month, day;
  int rv;
  GWEN_TYPE_UINT32 fromYmd;
  GWEN_TYPE_UINT32 toYmd;
  GWEN_IDLIST *idl;

  if (fromDate) {
    if (GWEN_Time_GetBrokenDownUtcDate(fromDate,
                                       &day, &month, &year)) {
      DBG_ERROR(0, "Error breaking down date");
      return false;
    }
    fromYmd=(year<<16)+((month+1)<<8)+day;
  }
  else
    fromYmd=0;

  if (toDate) {
    if (GWEN_Time_GetBrokenDownUtcDate(toDate,
                                       &day, &month, &year)) {
      DBG_ERROR(0, "Error breaking down date");
      return false;
    }
    toYmd=(year<<16)+((month+1)<<8)+day;
  }
  else
    toYmd=0xffffffff;

  DBG_NOTICE(0, "Loading transactions");
  if (!getStorage())
    return false;

  idl=AH_Storage_GetAvailableDays(_storage);
  if (!idl) {
    DBG_ERROR(0, "Error building storage index");
    return false;
  }
  rv=AH_Storage_GetFirstDay(_storage, idl, &year, &month, &day);
  while(!rv) {
    GWEN_TYPE_UINT32 id;
    GWEN_DB_NODE *db;
    GWEN_TYPE_UINT32 cymd;

    cymd=(year<<16)+(month<<8)+day;
    if (cymd>=fromYmd && cymd<=toYmd) {
      DBG_DEBUG(0, "Loading day %04d/%02d/%02d",
                year, month, day);

      id=AH_Storage_OpenDay(_storage, year, month, day, 1);
      if (!id) {
        DBG_ERROR(0, "Error loading day %04d/%02d/%02d",
                  year, month, day);
        GWEN_IdList_free(idl);
        return false;
      }

      db=AH_Storage_GetFirstTransaction(_storage, id);
      if (!db) {
        DBG_WARN(0, "No transactions in day %04d/%02d/%02d",
                 year, month, day);
      }
      while(db) {
        Transaction t;

        if (!t.fromDb(db)) {
          DBG_ERROR(0, "Bad format of transaction in day %04d/%02d/%02d",
                    year, month, day);
          AH_Storage_AbandonDay(_storage, id);
          GWEN_IdList_free(idl);
          return false;
        }
        DBG_ERROR(0, "Got a transaction");
        tl.push_back(t);
        db=AH_Storage_GetNextTransaction(_storage, id);
      }
      if (AH_Storage_CloseDay(_storage, id)) {
        DBG_ERROR(0, "Error closing day %04d/%02d/%02d",
                  year, month, day);
        GWEN_IdList_free(idl);
        return false;
      }
    }
    rv=AH_Storage_GetNextDay(_storage, idl, &year, &month, &day);
    if (rv) {
      DBG_NOTICE(0, "End of day list.");
    }
  } /* if first day */
  GWEN_IdList_free(idl);

  return true;
}



bool Account::loadTransactions(GWEN_TIME *fromDate,
                               GWEN_TIME *toDate,
                               std::list<RefPointer<Transaction> > &tl) {
  int year, month, day;
  int rv;
  GWEN_TYPE_UINT32 fromYmd;
  GWEN_TYPE_UINT32 toYmd;
  GWEN_IDLIST *idl;

  if (fromDate) {
    if (GWEN_Time_GetBrokenDownUtcDate(fromDate,
                                       &day, &month, &year)) {
      DBG_ERROR(0, "Error breaking down date");
      return false;
    }
    fromYmd=(year<<16)+((month+1)<<8)+day;
  }
  else
    fromYmd=0;

  if (toDate) {
    if (GWEN_Time_GetBrokenDownUtcDate(toDate,
                                       &day, &month, &year)) {
      DBG_ERROR(0, "Error breaking down date");
      return false;
    }
    toYmd=(year<<16)+((month+1)<<8)+day;
  }
  else
    toYmd=0xffffffff;

  DBG_NOTICE(0, "Loading transactions");
  if (!getStorage())
    return false;

  idl=AH_Storage_GetAvailableDays(_storage);
  if (!idl) {
    DBG_ERROR(0, "Error building storage index");
    return false;
  }

  rv=AH_Storage_GetFirstDay(_storage, idl, &year, &month, &day);
  while(!rv) {
    GWEN_TYPE_UINT32 id;
    GWEN_DB_NODE *db;
    GWEN_TYPE_UINT32 cymd;

    cymd=(year<<16)+(month<<8)+day;
    if (cymd>=fromYmd && cymd<=toYmd) {
      DBG_DEBUG(0, "Loading day %04d/%02d/%02d",
                year, month, day);

      id=AH_Storage_OpenDay(_storage, year, month, day, 1);
      if (!id) {
        DBG_ERROR(0, "Error loading day %04d/%02d/%02d",
                  year, month, day);
        GWEN_IdList_free(idl);
        return false;
      }

      db=AH_Storage_GetFirstTransaction(_storage, id);
      if (!db) {
        DBG_WARN(0, "No transactions in day %04d/%02d/%02d",
                 year, month, day);
      }
      while(db) {
        RefPointer<Transaction> t;

        t=new Transaction();
        if (!t.ref().fromDb(db)) {
          DBG_ERROR(0, "Bad format of transaction in day %04d/%02d/%02d",
                    year, month, day);
          AH_Storage_AbandonDay(_storage, id);
          GWEN_IdList_free(idl);
          return false;
        }
        DBG_ERROR(0, "Got a transaction");
        tl.push_back(t);
        db=AH_Storage_GetNextTransaction(_storage, id);
      }
      if (AH_Storage_CloseDay(_storage, id)) {
        DBG_ERROR(0, "Error closing day %04d/%02d/%02d",
                  year, month, day);
        GWEN_IdList_free(idl);
        return false;
      }
    }
    rv=AH_Storage_GetNextDay(_storage, idl, &year, &month, &day);
    if (rv) {
      DBG_NOTICE(0, "End of day list.");
    }
  } /* if first day */
  GWEN_IdList_free(idl);

  return true;
}



App *Account::getApi() const {
  return _banking;
}



AH_STORAGE *Account::getStorage(){
  if (_storage==0) {
    GWEN_BUFFER *pbuf;

    if (_bankCode.empty() || _accountNumber.empty()){
      DBG_ERROR(0, "Bank code AND account number needed");
      return false;
    }

    pbuf=GWEN_Buffer_new(0, 256, 0, 1);
    if (_banking->getAppUserDataDir(pbuf)) {
      DBG_ERROR(0, "Error getting data path");
      GWEN_Buffer_free(pbuf);
      return false;
    }
    GWEN_Buffer_AppendString(pbuf, "/transactions/280/");
    GWEN_Buffer_AppendString(pbuf, _bankCode.c_str());
    GWEN_Buffer_AppendString(pbuf, "/");
    GWEN_Buffer_AppendString(pbuf, _accountNumber.c_str());
    _storage=AH_Storage_new(GWEN_Buffer_GetStart(pbuf));
    GWEN_Buffer_free(pbuf);
  }
  assert(_storage);
  return _storage;
}



void Account::infoTransactionsChanged(){
  _transactions.clear();
}



void Account::setLastTransactionDate(const GWEN_TIME *t){
  if (_lastTransactionDate && t) {
    if (GWEN_Time_Diff(t, _lastTransactionDate)>0) {
      GWEN_Time_free(_lastTransactionDate);
      _lastTransactionDate=GWEN_Time_dup(t);
    }
  }
  else {
    GWEN_Time_free(_lastTransactionDate);
    if (t) _lastTransactionDate=GWEN_Time_dup(t);
    else _lastTransactionDate=0;
  }
}



const GWEN_TIME *Account::getLastTransactionDate() const{
  return _lastTransactionDate;
}



bool Account::saveTransaction(const Transaction *t) {
  const GWEN_TIME *ti;
  int year, month, day;
  GWEN_TYPE_UINT32 id;
  GWEN_TYPE_UINT32 tid;
  GWEN_DB_NODE *dbTransaction;
  GWEN_DB_NODE *dbT;

  dbTransaction=GWEN_DB_Group_new("transaction");
  if (!t->toDb(dbTransaction)) {
    DBG_ERROR(0, "Could not store transaction to db");
    GWEN_DB_Group_free(dbTransaction);
    return false;
  }
  ti=t->getDate();
  if (!ti)
    ti=t->getValutaDate();
  if (!ti) {
    DBG_ERROR(0, "No date in transaction, can not save");
    GWEN_DB_Group_free(dbTransaction);
    return false;
  }
  if (GWEN_Time_GetBrokenDownUtcDate(ti, &day, &month, &year)) {
    DBG_ERROR(0, "Bad date in transaction, can not save");
    GWEN_DB_Group_free(dbTransaction);
    return false;
  }

  id=AH_Storage_OpenDay(_storage, year, month+1, day, 0);
  if (!id) {
    DBG_ERROR(0, "Error loading day %04d/%02d/%02d",
              year, month, day);
    GWEN_DB_Group_free(dbTransaction);
    return false;
  }

  tid=t->getTransactionId();
  assert(tid);
  dbT=AH_Storage_GetFirstTransaction(_storage, id);
  while(dbT) {
    GWEN_TYPE_UINT32 i;

    DBG_ERROR(0, "Checking this transaction:");
    GWEN_DB_Dump(dbT, stderr, 2);
    i=GWEN_DB_GetIntValue(dbT, "transactionId", 0, -1);
    if (tid==i)
      break;
    dbT=AH_Storage_GetNextTransaction(_storage, id);
  }
  if (dbT)
    GWEN_DB_ClearGroup(dbT, 0);

  if (!dbT) {
    DBG_ERROR(0, "Transaction not found, adding it");
    if (AH_Storage_AddTransaction(_storage, id, dbTransaction)) {
      GWEN_DB_Group_free(dbTransaction);
      AH_Storage_AbandonDay(_storage, id);
      return false;
    }
  }
  else {
    DBG_ERROR(0, "Transaction found, modifying");
    GWEN_DB_AddGroupChildren(dbT, dbTransaction);
    GWEN_DB_Dump(dbT, stderr, 2);
    if (AH_Storage_CloseDay(_storage, id)) {
      DBG_ERROR(0, "Could not close day %04d/%02d/%02d",
                year, month, day);
      GWEN_DB_Group_free(dbTransaction);
      return false;
    }
  }
  GWEN_DB_Group_free(dbTransaction);

  return true;
}



bool Account::removeTransaction(const Transaction *t) {
  const GWEN_TIME *ti;
  int year, month, day;
  GWEN_TYPE_UINT32 id;
  GWEN_TYPE_UINT32 tid;
  GWEN_DB_NODE *dbT;

  ti=t->getDate();
  if (!ti)
    ti=t->getValutaDate();
  if (!ti) {
    DBG_ERROR(0, "No date in transaction, can not remove");
    return false;
  }
  if (GWEN_Time_GetBrokenDownUtcDate(ti, &day, &month, &year)) {
    DBG_ERROR(0, "Bad date in transaction, can not remove");
    return false;
  }

  id=AH_Storage_OpenDay(_storage, year, month+1, day, 0);
  if (!id) {
    DBG_ERROR(0, "Error loading day %04d/%02d/%02d",
              year, month, day);
    return false;
  }

  tid=t->getTransactionId();
  assert(tid);
  dbT=AH_Storage_GetFirstTransaction(_storage, id);
  while(dbT) {
    GWEN_TYPE_UINT32 i;

    i=GWEN_DB_GetIntValue(dbT, "transactionId", 0, -1);
    if (tid==i)
      break;
    dbT=AH_Storage_GetNextTransaction(_storage, id);
  }

  if (dbT) {
    DBG_ERROR(0, "Transaction found, removing");
    GWEN_DB_UnlinkGroup(dbT);
    GWEN_DB_Group_free(dbT);
    if (AH_Storage_CloseDay(_storage, id)) {
      DBG_ERROR(0, "Could not close day %04d/%02d/%02d",
                year, month, day);
      return false;
    }
  }

  return true;
}



bool Account::saveTransactions() {
  return true;
}











